#define CUDAFILE
#include "main.h"

void run_cpu_job(char *output, int threadid, task_data *task)
{
    /* Start Initialize the task*/
    int ioff_e, ioff_h, calcmet = task->calcmet_used;

    my_float *Hx1D, *Hy1D, *Hz1D, *Ex1D, *Ey1D, *Ez1D;
    my_float dt, Cbdx, Cbdy, Cbdz, Dbdx, Dbdy, Dbdz, p_source_factor;

    Hx1D = task->hx;
    Hy1D = task->hy;
    Hz1D = task->hz;
    Ex1D = task->ex;
    Ey1D = task->ey;
    Ez1D = task->ez;

    /* Initialize leap-frog update */
    dt = sqrt(3.0) / (2.0*c0*sqrt( 1.0/(task->dx*task->dx)+1.0/(task->dy*task->dy)+1.0/(task->dz*task->dz)));
    Cbdx = dt/(mu0*task->dx);
    Cbdy = dt/(mu0*task->dy);
    Cbdz = dt/(mu0*task->dz);
    Dbdx = dt/(eps0*task->dx);
    Dbdy = dt/(eps0*task->dy);
    Dbdz = dt/(eps0*task->dz);
    /* Initialize point source */
    p_source_factor = dt/(eps0*task->dx*task->dy*task->dz)*Const*2.0;
    int is = task->nx/2-1;
    int js = task->ny/2-1;
    int ks = task->nz/2-1;
    int nx = task->nx;
    int ny = task->ny;
    int nz = task->nz;

    memset(task->hx, 0, nx*ny*nz*sizeof(my_float));
    memset(task->hy, 0, nx*ny*nz*sizeof(my_float));
    memset(task->hz, 0, nx*ny*nz*sizeof(my_float));
    memset(task->ex, 0, (nx+1)*(ny+1)*(nz+1)*sizeof(my_float));
    memset(task->ey, 0, (nx+1)*(ny+1)*(nz+1)*sizeof(my_float));
    memset(task->ez, 0, (nx+1)*(ny+1)*(nz+1)*sizeof(my_float));


    ioff_e = (ny+1)*(nz+1);
    ioff_h = ny*nz;
    task->t[SETUP] += msecond();
    #ifdef verbose
    sprintf(output + strlen(output), "#%d# ====TASK=%2d=============CALCMET=5==========================\n", threadid, task->id);
    #endif
    double tf, perf[2];
    double   Flop, MFlops;
    perf[0] = 1.0e30;
    /* End Initialize the task*/

    for (int ts=1 ; ts<=task->nts ; ts++)
    {
        #ifdef verbose
        if (ts==1)
        {
            //printf(" ==========================================================\n");
            //printf(" Entering main loop\n");
            sprintf(output + strlen(output), "#%d# Total number of iterations are %i\n", threadid, task->nts);
        }
        #endif
        /* Yee_bench comment:
        Start timer. The first time step may differ from the others in
        performance, since it does not have any preceding time step. We do
        not want to pollute the timing by including it. */
        if (ts==2)
        {
            //printf(" Timer started at the beginning of the second time step\n");
            tf = -msecond();
        }
        
        /* OpenMP directive*/
        #pragma omp parallel num_threads(nomp)
        updateH_homo(calcmet,nx,ny,nz,ioff_e,ioff_h,Hx1D,Hy1D,Hz1D,
            Ex1D,Ey1D,Ez1D,Cbdx,Cbdy,Cbdz);

        /* Yee_bench comment:
         * In a real code there would be a lot of function calls in between
         * the calls to updateH_homo and updateE_homo. (which would alter
         * the value of the fields Hx, Hy, etc.)
         */
        #pragma omp parallel num_threads(nomp)
        updateE_homo(calcmet,nx,ny,nz,ioff_e,ioff_h,Hx1D,Hy1D,Hz1D,
            Ex1D,Ey1D,Ez1D,Dbdx,Dbdy,Dbdz);

        /* Yee_bench comment:
         * A point source in the center
         */
        Ez(is,js,ks) += - p_source_factor *( (ts*dt-3.0*Tk)/(Tk*Tk) )
            *exp( -1.0*(ts*dt-3.0*Tk)*(ts*dt-3.0*Tk)/(Tk*Tk) );
    
        /* Yee_bench comment: */
        /* PEC outer boundary conditions (F90 syntax) implicitly enforced: */
        /* Ey( 0  , :, :)=Ey(:, :,  0  )=Ey(nx+0, :, :)=Ey(:, :, nz+0)=0.0 */
        /* Ex(:,  0  , :)=Ex(:, :,  0  )=Ex(:, ny+0, :)=Ex(:, :, nz+0)=0.0 */
        /* Ez(:,  0  , :)=Ez( 0  , :, :)=Ez(:, ny+0, :)=Ez(nx+0, :, :)=0.0 */
        /* These fields are initialized to zero and stay zero.             */

    } /* Main timestepping loop end here */
    /* Do not insert anything here! It would disrupt the timing. */
    if (task->nts>1)
    {
        /* Timer not started if we only made one time step! */
        tf += msecond();
        Flop = (ny-1)*(nz-1) * 6 +  /*   Leftovers in calcmet = 5     */
            (nx-1)*(nz-1) * 6 +  /*   ------------ = ------------  */
            (nx-1)*(ny-1) * 6 +  /*   ------------ = ------------  */
            (double)(nx-1)*(double)(ny-1) * (double)(nz-1) * 18.0 +
            /*   The fused loops     */
            (double)nx*(double)ny*(double)nz * 18;
        MFlops = Flop*((double)task->nts-1.0)/(tf*1000000.0);
        #ifdef verbose
        sprintf(output + strlen(output), "#%d# -----------------------------------------------------\n", threadid);
        #else
        sprintf(output + strlen(output), "#%d# CPU Task %3d %3d ", threadid, task->id, task->nx);
        sprintf(output + strlen(output), "MFLOPS FOR MAINLOOP =  %8.2f MFlop/s. %6.2f sec.\n", MFlops, tf);
        #endif
        #ifdef verbose
        sprintf(output + strlen(output), "#%d#     MFLOPS   FOR MAINLOOP =  %6.2f MFlop/s.\n", threadid, MFlops);
        sprintf(output + strlen(output), "#%d#     USERTIME FOR MAINLOOP =  %6.2f sec.\n", threadid, tf);
        sprintf(output + strlen(output), "#%d# -----------------------------------------------------\n", threadid);
        #endif
        if (perf[0]>tf)
        {
            perf[0] = tf;
            perf[1] = MFlops;
        }
    } /* if nts>1 */
    #ifdef verbose
    sprintf(output + strlen(output), "#%d#     TIME  SPENT  SETUP    =  %8.3f sec.\n", threadid, task->t[SETUP]);
    #endif
    task->t[COMPUTE] = perf[0];
    task->mflops = perf[1];
    task->flop = Flop;
    
    task->t[COPY] = 0.0f;        
}

void run_gpu_job(char *output, int threadid, task_data *task, cudaStream_t stream)
{
    /* Start Initialize the task*/
    my_float *cHx, *cHy, *cHz, *cEx, *cEy, *cEz;
    my_float dt, Cbdx, Cbdy, Cbdz, Dbdx, Dbdy, Dbdz, p_source_factor; // midEz=0;
    double Flop, MFlops;

    int nx = task->nx;
    int ny = task->ny;
    int nz = task->nz;
    
    dim3 dimBlock(BLOCKX, BLOCKY); // max 512
    
    int grid_yz = ceil(sqrt(ny*nz/(float)(dimBlock.y*dimBlock.x)));
    int grid_xz = ceil(sqrt(nx*nz/(float)(dimBlock.y*dimBlock.x)));
    int grid_xy = ceil(sqrt(nx*ny/(float)(dimBlock.y*dimBlock.x)));
    
    dim3 dimGrid_yz(grid_yz, grid_yz);
    dim3 dimGrid_xz(grid_xz, grid_xz);
    dim3 dimGrid_xy(grid_xy, grid_xy);

    #ifdef verbose
    /* As the number of threads launched is almost never equal to the number of data points
     * over-saturation shows how many percent of the launched threads are immediately killed
     */
    sprintf(output + strlen(output), "#%d# G(%d,%d)xB(%d,%d) = %dvs%d -> %2.1f%%/%2.1f%%/%2.1f%% over-saturated\n",threadid , dimGrid_yz.x, dimGrid_yz.y, dimBlock.x, dimBlock.y, dimGrid_yz.x*dimGrid_yz.y * dimBlock.x*dimBlock.y, ny*nz, ((double)(dimGrid_yz.x*dimGrid_yz.y * dimBlock.x*dimBlock.y)/(ny*nz)-1)*100, ((double)(dimGrid_xz.x*dimGrid_xz.y * dimBlock.x*dimBlock.y)/(nx*nz)-1)*100,((double)(dimGrid_xy.x*dimGrid_xy.y * dimBlock.x*dimBlock.y)/(nx*ny)-1)*100);
    #endif

    /* Allocate all CUDA arrays to be used */
    CUDA_Success( cudaMalloc ( (void **)&cEx, (nx+1)*(ny+1)*(nz+1) * sizeof(my_float) ) );
    CUDA_Success( cudaMalloc ( (void **)&cEy, (nx+1)*(ny+1)*(nz+1) * sizeof(my_float) ) );
    CUDA_Success( cudaMalloc ( (void **)&cEz, (nx+1)*(ny+1)*(nz+1) * sizeof(my_float) ) );
    CUDA_Success( cudaMemset(cEx, 0,  (nx+1)*(ny+1)*(nz+1) * sizeof(my_float) ) );
    CUDA_Success( cudaMemset(cEy, 0,  (nx+1)*(ny+1)*(nz+1) * sizeof(my_float) ) );
    CUDA_Success( cudaMemset(cEz, 0,  (nx+1)*(ny+1)*(nz+1) * sizeof(my_float) ) );

    CUDA_Success( cudaMalloc ( (void **)&cHx, nx*ny*nz * sizeof(my_float) ) );
    CUDA_Success( cudaMalloc ( (void **)&cHy, nx*ny*nz * sizeof(my_float) ) );
    CUDA_Success( cudaMalloc ( (void **)&cHz, nx*ny*nz * sizeof(my_float) ) );
    CUDA_Success( cudaMemset(cHx, 0,  nx*ny*nz * sizeof(my_float) ) );
    CUDA_Success( cudaMemset(cHy, 0,  nx*ny*nz * sizeof(my_float) ) );
    CUDA_Success( cudaMemset(cHz, 0,  nx*ny*nz * sizeof(my_float) ) );
    
    /* CALCULATION START */
    
    /* Initialize leap-frog update */
    dt = sqrt(3.0) / (2.0*c0*sqrt( 1.0/(task->dx*task->dx)+1.0/(task->dy*task->dy)+1.0/(task->dz*task->dz)));
    Cbdx = dt/(mu0*task->dx);
    Cbdy = dt/(mu0*task->dy);
    Cbdz = dt/(mu0*task->dz);
    Dbdx = dt/(eps0*task->dx);
    Dbdy = dt/(eps0*task->dy);
    Dbdz = dt/(eps0*task->dz);
    /* Initialize point source */
    p_source_factor = dt/(eps0*task->dx*task->dy*task->dz)*Const*2.0;
    int is = task->nx/2-1;
    int js = task->ny/2-1;
    int ks = task->nz/2-1;

    /* End Initialize the task, sync the stream to get proper values for the setup timer
     * Else the stream would wait at the start of the first kernel instead.
     */
    CUDA_Success( cudaStreamSynchronize ( stream ), "syncsetup" );
    task->t[SETUP] += msecond ();

    double tf, perf[2];
    perf[0] = 1.0e30;

    for (int ts=1 ; ts<=task->nts ; ts++)
    {
        #ifdef verbose
        if (ts==1)
        {
            //" Entering main loop
            sprintf(output + strlen(output), "#%d#  Total number of iterations are %i\n", threadid, task->nts);
        }
        #endif

        /* Start timer.  */
        if (ts==2)
        {
            // Timer started at the beginning of the second time step
            tf = msecond();
        }
        CUDA_Success( cudaStreamSynchronize ( stream ), "sync1" );
        updateH_cuda<<<dimGrid_yz, dimBlock, 0, stream>>>(cHx, cHy, cHz, cEx, cEy, cEz, nx, ny, nz, Cbdx, Cbdy, Cbdz);

        /* Border */
        cleanupE_yz_cuda<<<dimGrid_yz, dimBlock, 0, stream>>>(cHy, cHz, cEx, nx, ny, nz, Dbdy, Dbdz);
        cleanupE_xz_cuda<<<dimGrid_xz, dimBlock, 0, stream>>>(cHx, cHz, cEy, nx, ny, nz, Dbdx, Dbdz);
        cleanupE_xy_cuda<<<dimGrid_xy, dimBlock, 0, stream>>>(cHx, cHy, cEz, nx, ny, nz, Dbdx, Dbdy);

        updateE_cuda<<<dimGrid_yz, dimBlock, 0, stream>>>(cHx, cHy, cHz, cEx, cEy, cEz, nx, ny, nz, Dbdx, Dbdy, Dbdz);

        CUDA_Success( cudaStreamSynchronize ( stream ), "sync2" );
        apply_source<<<1,1,0,stream>>>(cEz, p_source_factor, ts, dt, is, js, ks, ny, nz);
        
        /* Alternative to the single thread kernel above to apply source point. 
         * Copy data to host, calculate and copy back
         *

        CUDA_Success( cudaMemcpyAsync( &midEz, &cmEz(is,js,ks), sizeof(my_float), cudaMemcpyDeviceToHost, stream), "cpy to host" );
        CUDA_Success( cudaStreamSynchronize ( stream ), "sync3" );
        midEz += - p_source_factor *( (ts*dt-3.0*Tk)/(Tk*Tk) )
                 *exp( -1.0*(ts*dt-3.0*Tk)*(ts*dt-3.0*Tk)/(Tk*Tk) );
        CUDA_Success( cudaMemcpyAsync( &cmEz(is, js, ks), &midEz, sizeof(my_float), cudaMemcpyHostToDevice, stream), "cpy to device");
        
        * End alternative source point.
        */

    } /* Main timestepping loop end here */

    /* Do not insert anything here! It would disrupt the timing. */
    if (task->nts>1)
    {
        /* Timer not started if we only made one time step! */
        tf = msecond()-tf;
        Flop = (ny-1)*(nz-1) * 6 +  /*   Leftovers in calcmet = 5     */
            (nx-1)*(nz-1) * 6 +  /*   ------------ = ------------  */
            (nx-1)*(ny-1) * 6 +  /*   ------------ = ------------  */
            (double)(nx-1)*(double)(ny-1) * (double)(nz-1) * 18.0 +
            /*   The fused loops     */
            (double)nx*(double)ny*(double)nz * 18;
        MFlops = Flop*((double)task->nts-1.0)/(tf*1000000.0);
        #ifdef verbose
        sprintf(output + strlen(output), "#%d# -----------------------------------------------------\n", threadid);
        #else
        sprintf(output + strlen(output), "#%d# GPU Task %3d %3d ", threadid, task->id, task->nx);
        sprintf(output + strlen(output), "MFLOPS FOR MAINLOOP =  %8.2f MFlop/s. %6.2f sec.\n", MFlops, tf);
        #endif
        #ifdef verbose
        sprintf(output + strlen(output), "#%d#     MFLOPS   FOR MAINLOOP =  %6.2f MFlop/s.\n", threadid, MFlops);
        sprintf(output + strlen(output), "#%d#     USERTIME FOR MAINLOOP =  %6.2f sec.\n", threadid, tf);
        sprintf(output + strlen(output), "#%d# -----------------------------------------------------\n", threadid);
        #endif
        if (perf[0]>tf)
        {
            perf[0] = tf;
            perf[1] = MFlops;
        }
    } /* if nts>1 */
    
    /* gather data from GPU */
    task->t[COPY] = msecond();
    CUDA_Success( cudaMemcpyAsync ( task->ex, cEx, (nx+1)*(ny+1)*(nz+1)*sizeof(my_float), cudaMemcpyDeviceToHost, stream ) );
    CUDA_Success( cudaMemcpyAsync ( task->ey, cEy, (nx+1)*(ny+1)*(nz+1)*sizeof(my_float), cudaMemcpyDeviceToHost, stream ) );
    CUDA_Success( cudaMemcpyAsync ( task->ez, cEz, (nx+1)*(ny+1)*(nz+1)*sizeof(my_float), cudaMemcpyDeviceToHost, stream ) );
    CUDA_Success( cudaMemcpyAsync ( task->hx, cHx, nx*ny*nz*sizeof(my_float), cudaMemcpyDeviceToHost, stream ) );
    CUDA_Success( cudaMemcpyAsync ( task->hy, cHy, nx*ny*nz*sizeof(my_float), cudaMemcpyDeviceToHost, stream ) );
    CUDA_Success( cudaMemcpyAsync ( task->hz, cHz, nx*ny*nz*sizeof(my_float), cudaMemcpyDeviceToHost, stream ) );
    task->t[COPY] = msecond()-task->t[COPY];
    #ifdef verbose
    sprintf(output + strlen(output), "#%d#     TIME  SPENT  SETUP    =  %8.3f sec.\n", threadid, task->t[SETUP]);
    sprintf(output + strlen(output), "#%d#     TIME  SPENT  COPYING  =  %8.3f sec. %6.2fMb/s.\n", threadid, task->t[COPY], task->bytes/(task->t[COPY]*1024*1024));
    sprintf(output + strlen(output), "#%d#     IMPACT                =  %6.2f%%\n", threadid, (task->t[SETUP]+task->t[COPY])/perf[0]*100);
    sprintf(output + strlen(output), "#%d# -----------------------------------------------------\n", threadid);
    #endif

    /* write back results to task struct */
    task->t[COMPUTE] = perf[0];
    task->mflops = perf[1];
    task->flop = Flop;

    /* Free the CUDA arrays */
    cudaFree(cHx);
    cudaFree(cHy);
    cudaFree(cHz);
    cudaFree(cEx);
    cudaFree(cEy);
    cudaFree(cEz);
}
