#define CUDAFILE
#include "main.h"

int bytes_per_float;
//int d = 0;

/* Resources */
int ncpu, nomp, ngpu, nthreads;

/* Pthread Mutexes */
pthread_mutex_t pmutex;
pthread_mutex_t tmutex;
pthread_mutex_t hostmutex;

/* Greedy scheduler relative speed */
int fast=0;

/* Number of tasks we have*/
int number_of_tasks;

/* The array for our tasks */
task_data *tasks;
/* The id of the next task available */
int next_task = 0;

/* Runtime parameters
 * Which version are we running,
 * relates to enum in header
 */
int state = GREEDY;

/* Hybrid plotter */
int cpu_rested = 0;
int internal_tasks = 100;

/* Assignment scheduler */
double *workerperf; /* performance for each worker*/
int *workerassignment; /* ids of assigned tasks for each worker, see header file */
int requesting_worker = -1; /* id of worker that is currently communicating with scheduler */
int JOBS_PER_WORST_WORKER = 1; /* Old assignment scheme */
int MAX_TASKS_PER_THREAD = 100;
/* A worker will be assigned work to run for at least this long.  
 * Given that the current information about performance is correct. */
int SECONDS_TARGET = 20;

int main(int argc, char **argv)
{
	int err, nx, ny, nz, nts;
	my_float dx, dy, dz;
	FILE     *in_file;
	

#ifdef _WIN32
	StartCounter();
#endif
	/* ------------------------------------------------------------------------- */
	/*                    E x e c u t a b l e   c o d e                          */
	/* ------------------------------------------------------------------------- */

	printf(" This is the GPU-CPU Hybrid version of Yee_bench\n");
#ifdef use_float
	printf(" Using 32-bit precision (float)\n");
	bytes_per_float = 4;
#else
	printf(" Using 64-bit precision (double)\n");
	bytes_per_float = 8;
#endif

	/* Open, read and close yee.dat */
	in_file = fopen( "yee.dat", "r" );
	if( in_file == NULL )
	{
		printf("FATAL ERROR! Failed to open file yee.dat\n");
		exit(1);
	}
	err = fscanf(in_file, "%i %i %i%*[^\n]", &nx, &ny, &nz);
	printf(" nx, ny, nz = %i %i %i\n", nx, ny, nz);
#ifdef use_float
	err = fscanf(in_file, "%f %f %f%*[^\n]", &dx, &dy, &dz);
	printf(" dx, dy, dz = %f %f %f\n", dx, dy, dz);
#else
	err = fscanf(in_file, "%lf %lf %lf%*[^\n]", &dx, &dy, &dz);
	printf(" dx, dy, dz = %lf %lf %lf\n", dx, dy, dz);
#endif
	err = fscanf(in_file, "%i%*[^\n]", &nts);
	printf(" nts = %i\n", nts);

	err = fscanf(in_file, "%i%*[^\n]", &ncpu);
	printf(" ncpu = %i\n", ncpu);
	err = fscanf(in_file, "%i%*[^\n]", &nomp);
	printf(" nomp = %i\n", nomp);
	err = fscanf(in_file, "%i%*[^\n]", &ngpu);
	printf(" ngpu = %i\n", ngpu);
	err = fscanf(in_file, "%i", &number_of_tasks);
	printf(" number_of_tasks = %i\n", number_of_tasks);
	nthreads = ngpu + ncpu;
	/* compiler warning dummy */
	if (!err)
		printf("error\n");

	fclose( in_file );


	if (argc > 1)
	{
		if (strcmp(argv[1], "assignment") == 0)
			state = ASSIGNMENT;
		else if (strcmp(argv[1], "plot") == 0)
			state = PLOT;
		else if (strcmp(argv[1], "hybridplot") == 0)
			state = HPLOT;
		else if (strcmp(argv[1], "resource") == 0)
			state = RESOURCE;
		else
			state = GREEDY;
	}

	pthread_mutex_init(&pmutex, NULL);
	
	/* Start the correct function for which type of 
	 * mode we want run.
	 */
	if (state == PLOT)
		plot_single_device( nts, nx, ny, nz, dx, dy, dz);
	else if (state == HPLOT)
		plot_hybrid_setup( nts, nx, ny, nz, dx, dy, dz);
	else if (state == ASSIGNMENT)
		assignment_scheduler( nts, nx, ny, nz, dx, dy, dz);
	else if (state == RESOURCE)
		resource_bench( nts, nx, ny, nz, dx, dy, dz);
	else
		greedy_scheduler( nts, nx, ny, nz, dx, dy, dz);
	
	
	/* Allow the program to be run on a computer without CUDA */
	if (ngpu > 0)
		CUDA_Success( cudaDeviceReset(), "CUDA device reset" );

	pthread_mutex_destroy(&pmutex);

	/* Error check for CUDA implementation */
	#ifdef gpucheck
	int ioff_e = (ny+1)*(nz+1);
	fprintf(stderr, "Running error check\n");
	int cerr = 0;
	for (int ii=0; ii<nx; ii++)
	{
		for (int jj=0; jj<ny; jj++)
		{
			for (int kk=0; kk<nz; kk++)
			{
				if (abs(Ez1(ii,jj,kk) - Ez2(ii,jj,kk)) > 0.0001)
				{
					//printf("%f - %f\n", Ez1(ii,jj,kk), Ez2(ii,jj,kk));
					cerr++;
				}
			}
		}
	}
	printf(" cuda calc errors, %d/%d \n", cerr,nx*ny*nz);
	#endif

	printf("Run completed successfully\n");
	return 0;
}

void assignment_scheduler(int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz)
{
	/* Initialize scheduler arrays */
	tasks = (task_data*)calloc(number_of_tasks, sizeof(task_data));
	workerperf = (double*) calloc(nthreads, sizeof(double));
	workerassignment = (int*) calloc(nthreads * MAX_TASKS_PER_THREAD + nthreads, sizeof(int));

	double tot_time;
	pthread_mutex_init(&hostmutex, NULL);
	pthread_mutex_init(&tmutex, NULL);
	printf("Starting the assignment scheduler\n");

	/* Initialize all tasks with their size */
	for (int i = 0; i < number_of_tasks; i++)
	{
		tasks[i].nts = nts;
		
		tasks[i].nx = nx;
		tasks[i].ny = ny;
		tasks[i].nz = nz;

		tasks[i].dx = dx;
		tasks[i].dy = dy;
		tasks[i].dz = dz;
	}
	
	/* Lock the task mutex to wait for all threads to have been launched before starting timer */
	/* Mutex to stop worker threads to start before host is ready,
	 * and to make sure only one worker is communicating with the host at one time.
	 */
	pthread_mutex_lock (&pmutex);

	/* Mutex for host, will be unlocked by workers */
	pthread_mutex_lock (&hostmutex);

	/* Worker barrier when assigning new work */
	pthread_mutex_lock (&tmutex);

	pthread_t *pthreads = (pthread_t*) calloc(nthreads, sizeof(pthread_t));
	launch_threads(pthreads);

	/* Scheduler Start */

	/* Start workers and timer */
	pthread_mutex_unlock (&pmutex);
	tot_time = -msecond();


	while (true)
	{
		/* Wait for worker threads */
		pthread_mutex_lock (&hostmutex);
		/* if requesting_worker is -2 it means all work has been assigned, so end */
		if (requesting_worker == -2)
		{
			/* unlock to make sure any workers don't hang */
			pthread_mutex_unlock(&tmutex);	
			break;
		}
		assign_workload(requesting_worker);
		/* Signal the worker that the scheduler has done it's scheduling */
		pthread_mutex_unlock(&tmutex);
	}
	printf("#HOST# Ending assignments\n");

	/* Scheduler End */
	
	/* Join task threads when they are done */
	join_threads(pthreads);
	tot_time += msecond();

	// COMMON END

	printresults(nts, tot_time);
	pthread_mutex_destroy(&hostmutex);
	pthread_mutex_destroy(&tmutex);

}

void assign_workload(int worker_id)
{
	if (next_task >= number_of_tasks)
	{
		printf("#HOST# All assignments assigned already\n");
		/* Tell worker it has zero assignments to perform */
		TaskNum(worker_id) = 0;
		return;
	}
	/* Target flop for a worker to perform
	 * will try to make it as close to but at least
	 * SECONDS_TARGET seconds
	 */
	double flop_target = workerperf[worker_id] / (((double)tasks[next_task].nts-1.0)/(SECONDS_TARGET*1000000.0));
	printf("target for %d is %f\n", worker_id, flop_target/1000000);
	double flopassigned = 0;
	int tasks_assigned = 0;

	while (flopassigned < flop_target && tasks_assigned < MAX_TASKS_PER_THREAD && next_task < number_of_tasks)
	{
		/* Assign task */
		TaskId(worker_id, tasks_assigned) = next_task++;
		/* Keep track of how much work there is in the tasks we have assigned so far */
		flopassigned += get_flop(tasks[TaskId(worker_id, tasks_assigned++)]);
		//printf("worker %d got job %d with %f flop\n", worker_id, TaskId(worker_id, tasks_assigned), get_flop(tasks[TaskId(worker_id, tasks_assigned)]));
	}
	/* How many tasks in the tasklist for a worker */
	TaskNum(worker_id) = tasks_assigned;
	printf("#HOST# Assigning %d jobs to worker %d\n", TaskNum(worker_id), worker_id );
}

/*
 * This scheduler replacement will just have the workers perform the next task when available
 */
void greedy_scheduler(int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz)
{
	tasks = (task_data*)calloc(number_of_tasks, sizeof(task_data));
	double tot_time;
	printf("Starting the greedy scheduler\n");
	/* Initialize all tasks */
	for (int i = 0; i < number_of_tasks; i++)
	{
		tasks[i].nts = nts;
		
		tasks[i].nx = nx;
		tasks[i].ny = ny;
		tasks[i].nz = nz;

		tasks[i].dx = dx;
		tasks[i].dy = dy;
		tasks[i].dz = dz;
	}
	
	/* Lock the task mutex to wait for all threads to have been launched before starting timer */
	pthread_mutex_lock (&pmutex);
	pthread_t *pthreads = (pthread_t*) calloc(nthreads, sizeof(pthread_t));
	launch_threads(pthreads);

	/* Unlock and start timer */
	pthread_mutex_unlock (&pmutex);
	tot_time = -msecond();
	
	/* Join task threads when they are done */
	join_threads(pthreads);
	tot_time += msecond();

	printresults(nts, tot_time);
	free(pthreads);
}

/*
 * Plot performance for a single worker thread
 */
void plot_single_device(int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz)
{
	/* First N to run benchmark on */
	int startn = 50;

	tasks = (task_data*)calloc(number_of_tasks, sizeof(task_data));
	printf("Starting plot performance for N in range %d to %d\n", startn, nx);
	
	int dn = max((nx-startn)/number_of_tasks, 1);
	/* Generate tasks with their own size */
	for (int i = 0; i < number_of_tasks; i++)
	{
		tasks[i].nts = nts;
		int n = startn + i*dn;
		tasks[i].nx = n;
		tasks[i].ny = n;
		tasks[i].nz = nz; /* Will use one value for all Z */

		tasks[i].dx = dx;
		tasks[i].dy = dy;
		tasks[i].dz = dz;
	}

	/* Lock the task mutex to wait for all threads to have been launched before starting timer */
	pthread_mutex_lock (&pmutex);
	
	pthread_t *pthreads = (pthread_t*) calloc(nthreads, sizeof(pthread_t));
	launch_threads(pthreads);

	/* Unlock */
	pthread_mutex_unlock (&pmutex);
	
	/* Join task threads when they are done */
	join_threads(pthreads);

	/* Write results to csv file */
	FILE *plotfile;
	plotfile = fopen("data.csv", "w");
	if (plotfile == NULL)
	{
		printf("FATAL ERROR! Failed to open file data.csv\n");
		exit(1);
	}
	char row[256];
	printf("Writing data to CSV file\n");
	for (int i = 0; i < number_of_tasks; i++)
	{
		/* Uncomment next line to get Main Loop performance instead
		 * Comment out the other sprintf
		 */

		/* main loop */
		// sprintf(row, "%d\t%f\n", tasks[i].nx, tasks[i].mflops);
		
		/* total time spent */
		sprintf(row, "%d\t%f\n", tasks[i].nx, tasks[i].flop*((double)tasks[i].nts)/(tasks[i].t[TOTAL]*1000000.0));
		
		fputs(row, plotfile);
	}
	fclose(plotfile);
	free(pthreads);
}

/*
 * Plot performance for a any number of worker threads
 */
void plot_hybrid_setup(int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz)
{
	printf("Starting plot for hybrid setup sequence\n");
	int startn = 50;
	int dn = max((nx-startn)/number_of_tasks, 1);

	double *results = (double*)calloc(number_of_tasks * CSVFIELDS, sizeof(double));
	for (int k = 0; k < number_of_tasks; k++)
	{
		
		int n = startn + k*dn;
		double run_time = 0;
		tasks = (task_data*)calloc(internal_tasks, sizeof(task_data));
		for (int j = 0; j < internal_tasks; j++)
		{
			tasks[j].nts = nts;
			
			tasks[j].nx = n;
			tasks[j].ny = n;
			tasks[j].nz = nz;

			tasks[j].dx = dx;
			tasks[j].dy = dy;
			tasks[j].dz = dz;
		}

		/* Lock the task mutex to wait for all threads to have been launched before starting timer */
		pthread_mutex_lock (&pmutex);
		
		pthread_t *pthreads = (pthread_t*) calloc(nthreads, sizeof(pthread_t));
		launch_threads(pthreads);

		/* Unlock and start timer */
		pthread_mutex_unlock (&pmutex);
		run_time = -msecond();

		/* Join task threads when they are done */
		join_threads(pthreads);
		run_time += msecond();

		/* Reset values for next run */
		next_task = 0;
		cpu_rested = 0;
		fast = 0;
		
		double mflops;
		double flop = 0;
		double est_error;

		/* Sum up the amount of flop performed */
		for (int j=0 ; j < internal_tasks ; j++ )
		{
			if (tasks[j].done != 1)
				break;
			flop += tasks[j].flop;
		}

		int parsed = 0;
		enddata *ending_data = (enddata*) calloc(nthreads, sizeof(enddata));
		double extraflop = 0;
		for (int j = 0; j < nthreads; j++)
		{
			ending_data[j].calcmet = 0;
		}

		/* Collect data from the last task performed by each worker thread */
		for (int j=internal_tasks-1; j >= 0 ; j-- )
		{
			/* We start at the end, if it hasn't been performed we look at the next */
			if (tasks[j].done != 1)
				continue;

			/* Find worker id of the task, if we have not seen this worker before record it*/
			int id = tasks[j].workerid;
			if (ending_data[id].calcmet == 0)
			{
				ending_data[id].calcmet = tasks[j].calcmet_used;
				ending_data[id].flop = tasks[j].flop;
				ending_data[id].tot_time = tasks[j].t[TOTAL];
				ending_data[id].end_time = tasks[j].t[END];

				if (++parsed == nthreads)
					break;
			}
		}

		/*
		 * Find out how much of the time a worker did not perform any work.
		 * Calculate error estimation for performance.
		 * Emulate that a worker continued to work until the end.
		 * As other workers have more resources to use now we downplay it
		 * Especially CUDA threads which have calcmet 7.
		 * 
		 * Sort data of the last tasks performed with the oldest first
		 */
		qsort(ending_data, nthreads, sizeof(enddata), cmp_enddata);

		/* How much of this task is expected to finish at the same speed without
		 * affecting the speed of the other tasks in a positive manner
		 */
		double modifier;
		for (int j = 0; j < nthreads - 1; j++)
		{
			if (ending_data[j].calcmet == 7)
					modifier = 0.4;
				else
					modifier = 1.0;
			// Difference in time from when it ended vs the entire iteration
			double diff = ( ending_data[nthreads-1].end_time - ending_data[j].end_time ) / ending_data[j].tot_time;
			extraflop += diff * ending_data[j].flop * modifier;
		}

		mflops = (extraflop/2+flop)*((double)nts)/(run_time*1000000.0);
		est_error = extraflop/2*((double)nts)/(run_time*1000000.0);
		results[k * CSVFIELDS + CSVN] = n;
		results[k * CSVFIELDS + CSVMFLOPS] = mflops;
		results[k * CSVFIELDS + CSVERR] = est_error;
		printf("iteration %d, n=%d, mflops %f, esterr %f \n", k, n, mflops, est_error);
		free(tasks);
		free(ending_data);
		free(pthreads);
	}

	/* Write to CSV file */
	FILE *plotfile;
	plotfile = fopen("data.csv", "w");
	if (plotfile == NULL)
	{
		printf("FATAL ERROR! Failed to open file data.csv\n");
		exit(1);
	}
	char row[256];
	printf("Writing data to CSV file\n");
	for (int i = 0; i < number_of_tasks; i++)
	{
		sprintf(row, "%d\t%f\t%f\n", int(results[i*CSVFIELDS + CSVN]), results[i*CSVFIELDS + CSVMFLOPS], results[i*CSVFIELDS + CSVERR]);
		fputs(row, plotfile);
	}
	fclose(plotfile);
}

void resource_bench(int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz)
{
	printf("Starting resource benchmark\n");
	int max_cpu = 4;
	int max_gpu = 5;

	// make workers function as if its a hybrid plot
	state = HPLOT;
	
	/* Different setup for csv file in this one */
	/* 
	 * 1 nCPU
	 * 2 nGPU
	 * 3 MFlops
	 * 4 error
	 */
	int csvfields_r = 4;
	int entries = (max_cpu+1)*(max_gpu+1);

	double *results = (double*)calloc(entries * csvfields_r, sizeof(double));
	for (int cpu = 0; cpu <= max_cpu; cpu++)
	{
		for (int gpu = 0; gpu <= max_gpu; gpu++)
		{
			nomp = cpu;
			ngpu = gpu;

			if (cpu == 0 && gpu == 0){
				results[0] = 0;
				results[1] = 0;
				results[2] = 0;
				results[3] = 0;
				continue;
			}
				
			if (cpu > 0)
			{
				ncpu = 1;
				nthreads = 1+gpu;
			}
			else
			{
				ncpu = 0;
				nthreads = gpu;
			}
			printf("Testing %d CPU %d GPU \n", cpu, gpu);
			
			
			double run_time = 0;

			tasks = (task_data*)calloc(internal_tasks, sizeof(task_data));
			for (int j = 0; j < internal_tasks; j++)
			{
				tasks[j].nts = nts;
				
				tasks[j].nx = nx;
				tasks[j].ny = ny;
				tasks[j].nz = nz;

				tasks[j].dx = dx;
				tasks[j].dy = dy;
				tasks[j].dz = dz;
			}

			/* Lock the task mutex to wait for all threads to have been launched before starting timer */
			pthread_mutex_lock (&pmutex);
			
			pthread_t *pthreads = (pthread_t*) calloc(nthreads, sizeof(pthread_t));
			launch_threads(pthreads);

			/* Unlock and start timer */
			pthread_mutex_unlock (&pmutex);
			run_time = -msecond();

			/* Join task threads when they are done */
			join_threads(pthreads);
			run_time += msecond();

			/* Reset values between runs */
			next_task = 0;
			cpu_rested = 0;
			fast = 0;
			
			/* Rest of this is the same as for hybrid plot */
			double mflops;
			double flop = 0;
			
			double est_error;
			for (int j=0 ; j < internal_tasks ; j++ )
			{
				if (tasks[j].done != 1)
					break;
				flop += tasks[j].flop;
			}

			int parsed = 0;
			enddata *ending_data = (enddata*) calloc(nthreads, sizeof(enddata));
			double extraflop = 0;
			for (int j = 0; j < nthreads; j++)
			{
				ending_data[j].calcmet = 0;
			}
			for (int j=internal_tasks-1; j >= 0 ; j-- )
			{
				int id = tasks[j].workerid;
				if (tasks[j].done != 1)
					continue;

				if (ending_data[id].calcmet == 0)
				{
					ending_data[id].calcmet = tasks[j].calcmet_used;
					ending_data[id].flop = tasks[j].flop;
					ending_data[id].tot_time = tasks[j].t[TOTAL];
					ending_data[id].end_time = tasks[j].t[END];

					if (++parsed == nthreads)
						break;
				}
			}
			// Sort data of the last tasks performed with the oldest first
			qsort(ending_data, nthreads, sizeof(enddata), cmp_enddata);
			
			/* How much of this task is expected to finish at the same speed without
			 * affecting the speed of the other tasks in a positive manner
			 */
			double modifier = 1.0;
			for (int j = 0; j < nthreads - 1; j++)
			{
				if (ending_data[j].calcmet == 7)
					modifier = 0.4;
				else
					modifier = 1.0;
				// Difference in time from when it ended compared to the entire iteration
				// divided by the time it took to complete this task
				double diff = ( ending_data[nthreads-1].end_time - ending_data[j].end_time ) / ending_data[j].tot_time;
				extraflop += diff * ending_data[j].flop * modifier;
			}

			mflops = (extraflop/2+flop)*((double)nts)/(run_time*1000000.0);
			est_error = extraflop/2*((double)nts)/(run_time*1000000.0);
			
			results[(cpu * (max_gpu+1) + gpu) * csvfields_r + 0] = cpu;
			results[(cpu * (max_gpu+1) + gpu) * csvfields_r + 1] = gpu;
			results[(cpu * (max_gpu+1) + gpu) * csvfields_r + 2] = mflops;
			results[(cpu * (max_gpu+1) + gpu) * csvfields_r + 3] = est_error;
			printf("cpu %d, gpu %d, mflops %f, esterr %f \n", cpu, gpu, mflops, est_error);
			free(tasks);
			free(ending_data);
			free(pthreads);
		}
	}

	/* Save to CSV file */
	FILE *plotfile;
	plotfile = fopen("data.csv", "w");
	if (plotfile == NULL)
	{
		printf("FATAL ERROR! Failed to open file data.csv\n");
		exit(1);
	}
	char row[256];
	printf("Writing data to CSV file\n");
	for (int i = 0; i < entries; i++)
	{
		sprintf(row, "%d\t%d\t%f\t%f\n", int(results[i*csvfields_r + 0]), int(results[i*csvfields_r + 1]), results[i*csvfields_r + 2], results[i*csvfields_r + 3]);
		fputs(row, plotfile);
	}
	fclose(plotfile);

}


void launch_threads(pthread_t *pthreads)
{
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	printf("#HOST# Starting Task Threads\n");
	

	/* Launch the CPU threads */
	for (long i = 0; i < ncpu; i++)
	{
		int rc = pthread_create(&pthreads[i], &attr, cpu_thread, (void *)i);
		if (rc) {
			printf("ERROR; return code from pthread_create() is %d\n", rc);
	 		exit(-1);
		}
	}
	//printf("#HOST# CPU-Threads started: %d\n", CPUTHRDS);

	if (ngpu > 0)
	{
		CUDA_Success(  cudaDeviceSetCacheConfig(cudaFuncCachePreferL1), "cacheset" );
		/* Launch the GPU threads */
		for (long i = ncpu; i < nthreads; i++)
		{
			int rc = pthread_create(&pthreads[i], &attr, gpu_thread, (void *)i);
			if (rc) {
	         	printf("ERROR; return code from pthread_create() is %d\n", rc);
	         	exit(-1);
	        }
		}
		//printf("#HOST# GPU-Threads started: %d\n", GPUTHRDS);
	}
	pthread_attr_destroy(&attr);
}

void join_threads(pthread_t *pthreads)
{
	int rc = 0;
	for(int i=0; i < nthreads; i++)
	    {
	    	void *status;
			rc = pthread_join(pthreads[i], &status);
			if (rc) {
	         	printf("ERROR; return code from pthread_join() is %d\n", rc);
	         	exit(-1);
	        }
	        //printf("#HOST# Thread %d joined\n", i);
		}
		printf("#HOST# All threads joined\n");
}

void *cpu_thread(void *arg)
{
	/* Number of CPU tasks to complete for each problem size for heterogeneous benchmarks */
	int plot_precision = 2;
	int threadid = (int)((long)arg);

	printf("#%d# CPU-Thread started\n", threadid);
	int tasks_performed = 0;
	double lastperf=-1;
	int my_task;

	#ifdef verbose
	char output[550];
	#else
	char output[80];
	#endif

	int tasks_assigned = 0;
	int tasks_performed_for_assignment = 0;

	/* If using assignment scheduler
	 * run quick assessment of worker performance
	 */
	if (state == ASSIGNMENT)
	{
		/* Init test */ 
		task_data testtask;
		
		testtask.nts = 40;

		testtask.nx = tasks[0].nx;
		testtask.ny = tasks[0].ny;
		testtask.nz = tasks[0].nz;
		testtask.dx = tasks[0].dx;
		testtask.dy = tasks[0].dy;
		testtask.dz = tasks[0].dz;

		testtask.calcmet_used = 5;

		init_task(&testtask);

		run_cpu_job(output, threadid, &testtask);

		lastperf = testtask.mflops;
		destroy_task_arrays(&testtask);
		/* end of init test */
		
	}
	while (1)
	{
		output[0] = '\0';

		/* If using assignment scheduler */
		if (state == ASSIGNMENT)
		{	
			/* If the worker has completed all the currently scheduled tasks */
			if (tasks_performed_for_assignment == tasks_assigned)
			{
				/* Get new assignment unless all tasks are done */
				if (requesting_worker != -2)
				{
					pthread_mutex_lock (&pmutex);

					// Send signal to host
					workerperf[threadid] = lastperf;
					requesting_worker = threadid;
					pthread_mutex_unlock (&hostmutex);
					
					// Barrier to wait for host to assign
					pthread_mutex_lock (&tmutex);

					tasks_assigned = TaskNum(threadid);
					tasks_performed_for_assignment = 0;
					pthread_mutex_unlock (&pmutex);
				}
				if (tasks_assigned == 0 || requesting_worker == -2)
				{
					printf("#%d# Thread ending as there are no more tasks\n#%d# Performed %d tasks in total\n\n", threadid, threadid, tasks_performed);
					requesting_worker = -2;
					pthread_mutex_unlock (&hostmutex);
					pthread_exit((void*) 0);
				}
				printf("#%d# Got assignment for %d new tasks\n", threadid, tasks_assigned);
			}
			/* Next task to perform */
			my_task = TaskId(threadid, tasks_performed_for_assignment);

		}
		else
		{
			/* If we are not using the assignment scheduler
			 * we take the next available task from the global list
			 */
			pthread_mutex_lock (&pmutex);
			if (lastperf > 0)
			{
				if (lastperf > fast)
					fast = lastperf;
				/* If there are not that many tasks left, CPU will end before GPU */
				if ( state == GREEDY && fast/lastperf-1 > (number_of_tasks - next_task)*0.95)
				{
					//printf("#%d# perf%d  fast%d ratio%d rem%d\n", threadid, lastperf, fast, fast/lastperf, (int)((number_of_tasks - next_task)*0.95));
					printf("#%d# Thread ending as number of remaining tasks compared to perf is low(%d)\n#%d# Performed %d tasks in total\n", threadid, number_of_tasks - next_task, threadid, tasks_performed);
					pthread_mutex_unlock (&pmutex);
					pthread_exit((void*) 0);
				}
			}
			/* Next task to perform */
			my_task = next_task++;
			pthread_mutex_unlock (&pmutex);
		}

		if (my_task >= number_of_tasks)
		{
			printf("#%d# Thread ending as there are no more tasks\n#%d# Performed %d tasks in total\n\n", threadid, threadid, tasks_performed);
			pthread_exit((void*) 0);
		}

		task_data *task = &tasks[my_task];
		
		task->id=my_task;
		task->workerid=threadid;
		

		if (init_task(task))
		{
			printf("#%d# FATAL ERROR! Error during task initialization\n", threadid);
			exit(1);
		}

		task->calcmet_used = 5;

		run_cpu_job(output, threadid, task);
		
		//lastperf = lastperf*0.25 + task->mflops*0.75;
		lastperf = task->mflops;
		
		#ifdef verbose
		my_float sumEz = 0.0;
		for (int ii=0; ii<task->nx; ii++)
			for (int jj=0; jj<task->ny; jj++)
				for (int kk=0; kk<task->nz; kk++)
					sumEz = sumEz + taskEz(ii,jj,kk);
		sprintf(output + strlen(output), "#%d#  Checksum(Ez) = %25.17E \n", threadid, sumEz);
		if (task->nx==100 && task->ny==100 && task->nz==100 && task->nts==200)
			sprintf(output + strlen(output), "#%d#  Expected value  -6.92134e-02\n", threadid);
		#endif

		#ifdef verbose
		sprintf(output + strlen(output), "#%d# lastperf = %f\n", threadid, lastperf);
		sprintf(output + strlen(output), "#%d# ========================END================================\n\n", threadid);
		#endif
		tasks[my_task].done = 1;
		
		tasks_performed++;
		tasks_performed_for_assignment++;
		
		#ifndef gpucheck
		destroy_task_arrays(task);
		#endif
		task->t[END] = msecond();
		task->t[TOTAL] += task->t[END];
		printf("%s", output);

		/* When running the heterogeneous benchmark we only run a set number of CPU tasks  */
		if (state == HPLOT)
		{
			if (tasks_performed == plot_precision)
			{
				pthread_mutex_lock (&pmutex);
				cpu_rested += 1;
				pthread_mutex_unlock (&pmutex);
				printf("#%d# CPU Thread ending after performing %d task(s) for plot\n", threadid, plot_precision);
				pthread_exit((void*) 0);
			}
		}
	}
}

void *gpu_thread(void *arg)
{
	int threadid = (int)((long)arg);

	printf("#%d# GPU-Thread started\n", threadid);
	int tasks_performed = 0, my_task;
	double lastperf = -1;

	/* CUDA setup */

	cudaSetDevice(0);
	#ifdef verbose
	cudaDeviceProp p;
	printf("#%d# ================CUDA=INIT===========\n", threadid);
	cudaGetDeviceProperties (&p, 0);
	printf("#%d# %s with compute capability: %d.%d\n", threadid, p.name, p.major, p.minor);
	printf("#%d# %luMB mem\n", threadid, p.totalGlobalMem/(1024*1024));
	#endif
	cudaStream_t stream;
	cudaStreamCreate(&stream);

	#ifdef verbose
	char output[800];
	#else
	char output[80];
	#endif
	
	int tasks_assigned = 0;
	int tasks_performed_for_assignment = 0;
	
	/* If using assignment scheduler
	 * run quick assessment of worker performance
	 */
	if (state == ASSIGNMENT)
	{
		/* Init test */ 
		task_data testtask;
		
		testtask.nts = 40;

		testtask.nx = tasks[0].nx;
		testtask.ny = tasks[0].ny;
		testtask.nz = tasks[0].nz;
		testtask.dx = tasks[0].dx;
		testtask.dy = tasks[0].dy;
		testtask.dz = tasks[0].dz;
		testtask.calcmet_used = 7;

		init_task(&testtask);

		run_gpu_job(output, threadid, &testtask, stream);

		lastperf = testtask.mflops;
		destroy_task_arrays(&testtask);
		/* end of init test */
		
	}

	while (1)
	{
		output[0] = '\0';

		/* If using assignment scheduler */
		if (state == ASSIGNMENT)
		{
			/* If the worker has completed all the currently scheduled tasks */
			if (tasks_performed_for_assignment == tasks_assigned)
			{
				/* Get new assignment unless all tasks are done */
				if (requesting_worker != -2)
				{
					pthread_mutex_lock (&pmutex);

					// Send signal to host
					workerperf[threadid] = lastperf;
					requesting_worker = threadid;
					pthread_mutex_unlock (&hostmutex);
					
					// Barrier to wait for host to assign
					pthread_mutex_lock (&tmutex);

					tasks_assigned = TaskNum(threadid);
					tasks_performed_for_assignment = 0;
					pthread_mutex_unlock (&pmutex);
				}
				if (tasks_assigned == 0 || requesting_worker == -2)
				{
					printf("#%d# Thread ending as there are no more tasks\n#%d# Performed %d tasks in total\n\n", threadid, threadid, tasks_performed);
					requesting_worker = -2;
					pthread_mutex_unlock (&hostmutex);
					pthread_exit((void*) 0);
				}
				printf("#%d# Got assignment for %d new tasks\n", threadid, tasks_assigned);
			}
			/* Next task to perform */
			my_task = TaskId(threadid, tasks_performed_for_assignment);

		}
		else
		{
			/* If we are not using the assignment scheduler
			 * we take the next available task from the global list
			 */
			pthread_mutex_lock (&pmutex);
			if (lastperf > fast)
				fast = lastperf;
			
			if (state == HPLOT && cpu_rested == ncpu && tasks_performed_for_assignment > 0)
			{
				pthread_mutex_unlock (&pmutex);
				printf("#%d# GPU Thread ending as all CPU tasks performed for plot\n#%d# Performed %d tasks in total\n", threadid, threadid, tasks_performed);
				pthread_exit((void*) 0);
			}
			/* Next task to perform */
			my_task = next_task++;
			pthread_mutex_unlock (&pmutex);
		}
		
		if (state == HPLOT && my_task >= internal_tasks)
		{
			printf("ERROR: RAN OUT OF TASKS WHILE PERFORMING HYBRID PLOT TEST\n");
			exit(1);
		}
		if (state != HPLOT && my_task >= number_of_tasks)
		{
			printf("#%d# GPU Thread ending as there are no more tasks\n#%d# Performed %d tasks in total\n", threadid, threadid, tasks_performed);
			pthread_exit((void*) 0);
		}
		task_data *task = &tasks[my_task];
		task->id = my_task;
		task->workerid=threadid;
		
		if (init_task(task))
		{
			printf("#%d# FATAL ERROR! Error during task initialization\n", threadid);
			exit(1);
		}
		
		/* Set calcmet_used to CUDA */
		task->calcmet_used = 7;
		
		#ifdef verbose
		sprintf(output + strlen(output), "#%d# ====TASK=%2d====CUDA=====CALCMET=7==========N=%3d==========\n", threadid, my_task, task->nx);
		#endif

		run_gpu_job(output, threadid, task, stream);

        //lastperf = lastperf*0.25 + task->mflops*0.75;
		lastperf = task->mflops;

		#ifdef verbose
		my_float sumEz = 0.0;
		for (int ii=0; ii<task->nx; ii++)
			for (int jj=0; jj<task->ny; jj++)
				for (int kk=0; kk<task->nz; kk++)
					sumEz = sumEz + taskEz(ii,jj,kk);
		sprintf(output + strlen(output), "#%d#  Checksum(Ez) = %25.17E \n", threadid, sumEz);
		if (task->nx==100 && task->ny==100 && task->nz==100 && task->nts==200)
			sprintf(output + strlen(output), "#%d# Expected value  -6.92134e-02\n", threadid);

		sprintf(output + strlen(output), "#%d# lastperf = %f\n", threadid, lastperf);
		sprintf(output + strlen(output), "#%d# ========================END===============================\n\n", threadid);
		#endif

		task->done = 1;
		tasks_performed++;
		tasks_performed_for_assignment++;
		#ifndef gpucheck
        destroy_task_arrays(task);
        #endif
        task->t[END] = msecond();
		task->t[TOTAL] += task->t[END];
		printf("%s", output);
	}
}

/* Helper function for how many flop for a task */
double get_flop(task_data task)
{
	int nx = task.nx;
	int ny = task.ny;
	int nz = task.nz;

	return (ny-1)*(nz-1) * 6 +  /*   Leftovers in calcmet = 5     */
                (nx-1)*(nz-1) * 6 +  /*   ------------ = ------------  */
                (nx-1)*(ny-1) * 6 +  /*   ------------ = ------------  */
                (double)(nx-1)*(double)(ny-1) * (double)(nz-1) * 18.0 +
                /*   The fused loops     */
                (double)nx*(double)ny*(double)nz * 18;
}

int destroy_task_arrays(task_data *task)
{
	free(task->ex);
	free(task->ey);
	free(task->ez);
	free(task->hx);
	free(task->hy);
	free(task->hz);
	return 0;
}

void printresults(int nts, double tot_time)
{

	double sumtime[TIMERS], sumtime2[TIMERS];
	double sumflop[] = {0, 0};
	memset(sumtime,  0, TIMERS*sizeof(double));
	memset(sumtime2, 0, TIMERS*sizeof(double));
	double flop = 0.0;
	int gputasks = 0, cputasks;
	
	
	for (int i = 0; i < number_of_tasks; i++)
	{
		flop += tasks[i].flop;
		if (tasks[i].calcmet_used == 7)
		{
			sumflop[0] += tasks[i].flop;
			for (int j = 0; j < TIMERS ; j++)
				sumtime[j] += tasks[i].t[j];
			gputasks++;
		}
		else
		{
			sumflop[1] += tasks[i].flop;
			for (int j = 0; j < TIMERS ; j++)
				sumtime2[j] += tasks[i].t[j];
		}
	}
	cputasks = number_of_tasks - gputasks;

	double mFlops = flop*((double)nts)/(tot_time*1000000.0);
	int m = int(tot_time/60);
	int s = int(tot_time - m*60);
	int ms = int((tot_time - m*60 - s)*1000);

	double avgmflops_g = sumflop[0] * ((double)nts)/(sumtime[TOTAL]*1000000.0);
	double avgmflops_c = sumflop[1] * ((double)nts)/(sumtime2[TOTAL]*1000000.0);

	printf("\n\n");
	printf("Total  Launch time = %8.3f sec. (%d:%02d:%03d)\n", tot_time, m, s, ms);
	printf("Total   MFLOPS     = %8.2f MFlops/s\n\n", mFlops);
	
	printf("------------------CUDA-TASKS-%3d-------------------\n", gputasks);
	printf("Setup       %8.3f sec.\n", sumtime[SETUP]/gputasks);
	printf("Compute     %8.3f sec.\n", sumtime[COMPUTE]/gputasks);
	printf("Copy        %8.3f sec.\n", sumtime[COPY]/gputasks);
	printf("Other       %8.3f sec.\n\n", (sumtime[TOTAL]-sumtime[COPY]-sumtime[COMPUTE]-sumtime[SETUP])/gputasks);
	printf("Average MFlops %f\n\n", avgmflops_g);
	
	printf("------------------CPU--TASKS-%3d---------------------\n", cputasks);
	printf("Setup       %8.3f sec.\n", sumtime2[SETUP]/cputasks);
	printf("Compute     %8.3f sec.\n", sumtime2[COMPUTE]/cputasks);
	printf("Copy        %8.3f sec.\n", sumtime2[COPY]/cputasks);
	printf("Other       %8.3f sec.\n\n", (sumtime2[TOTAL]-sumtime2[COPY]-sumtime2[COMPUTE]-sumtime2[SETUP])/cputasks);
	printf("Average MFlops %f\n\n", avgmflops_c);
}

int init_task(task_data *task)
{
	int nts = task->nts;
	int nx = task->nx;
	int ny = task->ny;
	int nz = task->nz;

	memset(task->t, 0, TIMERS*sizeof(double));
	task->t[SETUP] = -msecond ();
	task->t[TOTAL] = task->t[SETUP];

	if (nts <= 0) 
	{
		printf("nts is zero or less \n");
		return 1;
	}

	if (nx <= 0 || ny <= 0 || nz <= 0) 
	{
		printf("ERROR! a dimension is less than 0 \n");
		return 1;
	}

	/* Allocate memory for the electromagnetic fields. */
	task->ex = (my_float*)calloc((nx+1)*(ny+1)*(nz+1),sizeof(my_float));
	if (task->ex == NULL )
	{
		printf("FATAL ERROR! Failed to allocate Ex! \n");
		return(1);
	}

	task->ey = (my_float*)calloc((nx+1)*(ny+1)*(nz+1),sizeof(my_float));
	if (task->ey == NULL )
	{
		printf("FATAL ERROR! Failed to allocate Ey! \n");
		return(1);
	}

	task->ez = (my_float*)calloc((nx+1)*(ny+1)*(nz+1),sizeof(my_float));
	if (task->ez == NULL )
	{
		printf("FATAL ERROR! Failed to allocate Ez! \n");
		return(1);
	}

	task->hx = (my_float*)calloc(nx*ny*nz,sizeof(my_float));
	if (task->hx == NULL )
	{
		printf("FATAL ERROR! Failed to allocate Hx! \n");
		return(1);
	}

	task->hy = (my_float*)calloc(nx*ny*nz,sizeof(my_float));
	if (task->hy == NULL )
	{
		printf("FATAL ERROR! Failed to allocate Hy! \n");
		return(1);
	}

	task->hz = (my_float*)calloc(nx*ny*nz,sizeof(my_float));
	if (task->hz == NULL )
	{
		printf("FATAL ERROR! Failed to allocate Hz! \n");
		return(1);
	}


	task->bytes = 3.0*( (double)nx    *(double)ny    *(double)nz +
		(double)(nx+1)*(double)(ny+1)*(double)(nz+1) );
	task->bytes = task->bytes*(double)bytes_per_float;
	#ifdef verbose
	printf(" Electromagnetic fields allocated, bytes used =  %i\n", task->bytes);
	#endif

	return 0;
}

/* Walltime */
double msecond()
{
#ifdef _WIN32
	return GetCounter()/1000;
#else
	static struct timeval t;
	gettimeofday ( &t, NULL );
	return ( t.tv_sec + 1e-6 * t.tv_usec );
#endif
}

void CUDA_Success(cudaError_t error, char *str)
{
    if(error != cudaSuccess)
    {
        printf( "CUDA Error: %s - %s\n", cudaGetErrorString( error ), str );
        exit(EXIT_FAILURE);
    }
}

void CUDA_Success(cudaError_t error)
{
    CUDA_Success(error, "");
}

/* Sort enddata structs based on the timestamp for end */
int cmp_enddata(const void *a,const void *b) {
	enddata *x = (enddata *)a;
	enddata *y = (enddata *)b;
	if (x->end_time < y->end_time) return -1;
	else if (x->end_time > y->end_time) return 1;
	return 0;
}