#!/usr/bin/python
import csv
import os
import matplotlib
from scipy.interpolate import interp1d
import numpy as np

DEFAULTFILE = "data.csv"
DEFAULTFILETYPE = "png"

import argparse
parser = argparse.ArgumentParser(description='Plot CSV files for yee_bench')
parser.add_argument("-f", "--file", type=str, nargs='+', help="specify csv input filename, default is "+DEFAULTFILE, default=[DEFAULTFILE])
parser.add_argument("-a", "--all", action="store_true", help="plot all local csv files")
parser.add_argument("-d", "--dir", type=str, help="directory for all argument")
parser.add_argument("-s", "--save", action="store_true", help="save output to file, default is "+DEFAULTFILETYPE)
parser.add_argument("-t", "--filetype", type=str, help="file type for saved output", default=DEFAULTFILETYPE)
parser.add_argument("-v", "--verbose", action="store_true", help="be verbose")
parser.add_argument("-e", "--errorbars", action="store_true", help="include error bars")
parser.add_argument("-r", "--resource", action="store_true", help="test of resource allocation")
parser.add_argument("-n", "--firstn", type=int, help="start the graph at n")
parser.add_argument("-m", "--smoothing", action="count", help="graph smoothing, the  number of m signifying the number of consolidated data points", default=0)
args = parser.parse_args()

if args.save:
    matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import rc


def getColumn(filename, column):
    results = csv.reader(open(filename), delimiter="\t")
    if column == 0 or (column == 1 and args.resource):
        return [int(result[column]) for result in results]
    else:
        return [float(result[column]) for result in results]


def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in xrange(0, len(l), n):
        yield l[i+1:i+n]


def plotfile(filein):
    if args.verbose:
        print "plotting ", filein
    x = getColumn(filein, 0)
    y = getColumn(filein, 1)
    if args.firstn:
        y = y[x.index(args.firstn):]
        x = x[x.index(args.firstn):]

    startid = filein.rfind("/")+1 if ("/" in filein) else 0
    name = filein[startid:filein.index(".csv")].replace("_", "\_")

    # Graph smoothin, requires scipy
    if args.smoothing > 0:
        if args.verbose:
            print "Smoothing to ", args.smoothing
        x2 = np.linspace(min(x), max(x), len(x)/args.smoothing)
        y2 = [sum([y[i*args.smoothing + k] for k in range(args.smoothing)])/args.smoothing for i in range(len(y)/args.smoothing)]
        f = interp1d(x2, y2, kind='cubic')
        xn = np.linspace(min(x), max(x), len(x)*10)

    if args.errorbars:
        y_err = getColumn(filein, 2)
        plt.errorbar(x, y, yerr=y_err, label=name)
    else:
        if args.smoothing > 0:
            plt.plot(xn, f(xn), label=name)
        else:
            plt.plot(x, y, label=name)


def plotresourcefile(filein):
    if args.verbose:
        print "plotting resource", filein
    c = getColumn(filein, 0)
    g = getColumn(filein, 1)
    y = getColumn(filein, 2)
    y_err = getColumn(filein, 3)

    cpus = max(c) + 1
    gpus = max(g) + 1
    x2 = list(chunks(g, gpus))
    y2 = list(chunks(y, gpus))
    ye2 = list(chunks(y_err, gpus))
    if args.verbose:
        print "CPUs:", cpus, "GPUs:", gpus
    for i in range(cpus):
        name = str(i)+"CPU"
        x2[i] = [k+0.02*(i-cpus/2) for k in x2[i]]
        if args.errorbars:
            plt.errorbar(x2[i], y2[i], yerr=ye2[i], label=name)
        else:
            plt.plot(x2[i], y2[i], label=name)

rc('text', usetex=True)
rc('font', family='serif', serif='Computer Modern Roman', size=12)
if args.resource:
    plt.figure("Resource Allocation Performance")
    plt.xlabel("GPU Streams")
else:
    plt.figure("n to time")
    plt.xlabel("n")
plt.ylabel("mflops")
ax = plt.subplot(111)

files = []

if args.all:
    path = args.dir if args.dir else "."
    if args.verbose:
        print "Plotting dir", path
    for f in os.listdir(path):
        if f.endswith(".csv"):
            files.append(os.path.join(path, f))
    files.sort()
else:
    for f in args.file:
        if not os.path.exists(f):
            print "Error: Plotfile '" + f + "' does not exist"
            exit(1)
        files.append(f)

if len(files) == 0:
        print "No files found for plot"
        exit()

for f in files:
    if args.resource:
        plotresourcefile(f)
    else:
        plotfile(f)

if len(files) > 1 or args.resource:
    ax.legend(loc='best') #(loc='center left', bbox_to_anchor=(0.7, 0.2))

ax.grid(True, which='both')
plt.ylim(ymin=0000)
fig = matplotlib.pyplot.gcf()
Size = fig.get_size_inches()
fig.set_size_inches(8.0,4.0)
if args.verbose:
    print "Size in Inches", Size

if args.save:
    figure = matplotlib.figure.Figure()
    fcb = matplotlib.backends.backend_agg.FigureCanvasBase(figure)
    supported_file_types = fcb.get_supported_filetypes()
    ft = args.filetype.replace(".", "")
    if ft not in supported_file_types:
        print "Unsupported filetype: ", args.filetype
        exit(1)
    filename = 'plotfile.'+ft
    if args.verbose:
        print "saving to: " + filename
    plt.savefig(filename, bbox_inches='tight', dpi=100)
else:
    if args.verbose:
        print "showing plot"
    plt.show()
